/*-------------------------------------------------------------------------

			  Raw to PoV Converter
		     Copyright (c) 1992 Steve Anger

   Converts lists of triangle coordinates in raw ASCII text format to
 PoV files. Automatically adds bounding shapes and produces smooth triangles.
 This file may be freely modified and distributed.

                                      CompuServe: 70714,3113
					Internet: steve.anger@rose.com
			               YCCMR BBS: (708)358-5611

--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "rayopt.h"

#define VERSION "v1.41"
#define MAX_TEXTURE 110

#if defined (__TURBOC__) || defined (__BORLANDC__)
extern unsigned _stklen = 16384;
#endif

void process_args (int argc, char *argv[]);
int  parse_line (FILE *f);
void update_txtlist (char *new_texture);
void fswap (float *a, float *b);

char  infile[64];       /* Name of input file */
char  outfile[64];      /* Name of output file */
float smooth;           /* Smooth triangles with angles < this */
int   bound;            /* Type of bounding shape to use */
int   verbose;          /* Verbose status messages */
int   one_object;       /* Optimize file as a single object */
int   swap_yz;          /* Swap Y and Z coordinates */

char  *txtlist[MAX_TEXTURE];
int   txtcnt = 0;

float  ax, ay, az;
float  bx, by, bz;
float  cx, cy, cz;
float  red, green, blue;
char   new_obj_name[80];
char   new_texture[80];


int main (int argc, char *argv[])
{
    FILE  *f;
    float gmin_x, gmin_y, gmin_z, gmax_x, gmax_y, gmax_z;
    int   line_cnt, obj_cnt, tri_cnt;
    int   done, i;
    char  obj_name[80] = "";

    process_args (argc, argv);

    f = fopen (infile, "r");

    if (f == NULL) {
	printf ("Error opening input file %s\n", infile);
	exit(1);
    }

    opt_set_dec (4);
    opt_set_bound (bound);
    opt_set_smooth (smooth);
    opt_set_quiet (!verbose);
    opt_set_fname (outfile, "");

    /* Use the name of the file as default object name */
    strcpy (obj_name, infile);
    add_ext (obj_name, "", 1);

    line_cnt = 0;
    obj_cnt = 0;
    tri_cnt = 0;
    done = 0;

    if (one_object)
	printf ("Working on: %s\n", obj_name);

    while (!done) {
	switch (parse_line(f)) {
	    case -1: printf ("Error in input file, line %d\n", line_cnt+1);
		     exit(1);

	    case  0: done = 1;
		     break;

	    case  1: opt_set_color (1.0, 1.0, 1.0);
		     opt_add_tri (ax, ay, az, bx, by, bz, cx, cy, cz);
		     ++tri_cnt;
		     break;

	    case  2: opt_set_color (red, green, blue);
		     opt_add_tri (ax, ay, az, bx, by, bz, cx, cy, cz);
		     ++tri_cnt;
		     break;

	    case  3: update_txtlist (new_texture);

		     opt_set_texture (new_texture);
		     opt_add_tri (ax, ay, az, bx, by, bz, cx, cy, cz);
		     ++tri_cnt;
		     break;

	    case  4: if (!one_object) {
			 if (tri_cnt > 0) {
			     opt_write_pov (obj_name);
			     ++obj_cnt;
			 }

			 strcpy (obj_name, new_obj_name);
			 printf ("Working on: %s\n", obj_name);
		     }
		     break;
	}
    }

    opt_write_pov (obj_name);
    ++obj_cnt;

    printf ("\n");
    fclose (f);

    f = fopen (outfile, "a");
    if (f == NULL)
	abortmsg ("Error opening output file", 1);

    if (txtcnt > 0)
	fprintf (f, "\n/* Named textures, modify as needed */\n");

    for (i = 0; i < txtcnt; i++) {
	fprintf (f, "#declare %s = texture\n", txtlist[i]);
	fprintf (f, "    DefTexture\n");
	fprintf (f, "    color red 1.0 green 1.0 blue 1.0\n");
	fprintf (f, "end_texture\n\n");

	free (txtlist[i]);
    }

    fprintf (f, "composite   /* All Objects */\n    ");

    fclose (f);
    opt_finish();
    f = fopen (outfile, "a");

    if (f == NULL)
	abortmsg ("Error opening output file", 1);

    opt_get_glimits (&gmin_x, &gmin_y, &gmin_z, &gmax_x, &gmax_y, &gmax_z);

    if (obj_cnt > 2) {
	fprintf (f, "\n");
	fprintf (f, "    bounded_by\n");
	fprintf (f, "\tintersection\n");
	fprintf (f, "\t   plane <+1 0 0> %.4f end_plane\n", +gmax_x);
	fprintf (f, "\t   plane <-1 0 0> %.4f end_plane\n", -gmin_x);
	fprintf (f, "\t   plane <0 +1 0> %.4f end_plane\n", +gmax_y);
	fprintf (f, "\t   plane <0 -1 0> %.4f end_plane\n", -gmin_y);
	fprintf (f, "\t   plane <0 0 +1> %.4f end_plane\n", +gmax_z);
	fprintf (f, "\t   plane <0 0 -1> %.4f end_plane\n", -gmin_z);
	fprintf (f, "\tend_intersection\n");
	fprintf (f, "    end_bound\n");
    }

    fprintf (f, "\n");
    fprintf (f, "    /*\n");
    fprintf (f, "        Scene extents\n");
    fprintf (f, "        X - Min: %8.4f  Max: %8.4f\n", gmin_x, gmax_x);
    fprintf (f, "        Y - Min: %8.4f  Max: %8.4f\n", gmin_y, gmax_y);
    fprintf (f, "        Z - Min: %8.4f  Max: %8.4f\n", gmin_z, gmax_z);
    fprintf (f, "    */\n");

    fprintf (f, "end_composite\n\n");

    fclose(f);

    return 0;
}


void process_args (int argc, char *argv[])
{
    int i;

    printf ("\n");
    printf ("RAW2POV %s - Copyright (c) 1992 Steve Anger\n", VERSION);
#if defined(__GNUC__) && defined(i386)
    printf ("DOS Extender Copyright (c) 1991 DJ Delorie\n");
#endif
    printf ("This program is freely distributable\n");
    printf ("\n");

    if (argc < 2) {
	printf ("Usage: raw2pov inputfile[.raw] [outputfile[.dat]] [[-/]options]\n\n");
	printf ("Options: -snnn  - Smooth triangles with angles < nnn\n");
	printf ("         -b     - Use box as bounding shape instead of ellipsoid.\n");
	printf ("         -u     - Do not add nested bounds to output (unbounded)\n");
	printf ("         -v     - Verbose status messages\n");
	printf ("         -1     - Generate one object only\n");
	printf ("         -x     - Exchange the Y and Z coords\n");
	printf ("\nex. raw2pov chess.raw chess.dat -s60 -v\n\n");
	exit(1);
    }

    infile[0] = '\0';
    outfile[0] = '\0';
    smooth = 0.0;
    bound = 0;
    verbose = 0;
    one_object = 0;
    swap_yz = 0;

    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-' || argv[i][0] == '/') {
	    switch (argv[i][1]) {
		case 'B':
		case 'b': bound = 1;
			  break;

		case 'U':
		case 'u': bound = 2;
			  break;

		case 'V':
		case 'v': verbose = 1;
			  break;

		case '1': one_object = 1;
			  break;

		case 'X':
		case 'x': swap_yz = 1;
			  break;

		case 'S':
		case 's': if (argv[i][2] == '\0')
			      smooth = 60.0;
			  else
			      sscanf (&argv[i][2], "%f", &smooth);
			  break;

		default : printf ("\nInvalid option -%c\n", argv[i][1]);
			  exit (1);
	    }
	}
	else if (infile[0] == '\0') {
	    strcpy (infile, argv[i]);
	    add_ext (infile, "raw", 0);
	}
	else if (outfile[0] == '\0') {
	    strcpy (outfile, argv[i]);
	    add_ext (outfile, "dat", 0);
	}
	else
	    abortmsg ("Too many file names.\n", 1);
    }

    if (outfile[0] == '\0') {
	strcpy (outfile, infile);
	add_ext (outfile, "dat", 1);
    }
}



int parse_line (FILE *f)
{
    char line[256], tokens[12][64], *p, *q;
    int  tcnt, n, i;

    if (fgets (line, 256, f) == NULL)
	return 0;

    line[255] = '\0';
    tcnt = 0;

    p = line;
    while (*p != '\0') {
	/* Skip white space */
	while (isspace(*p))
	    ++p;

	if (tcnt == 12  &&  *p != '\0')
	    return -1;

	/* Read next token */
	q = tokens[tcnt];
	while (!isspace(*p)  &&  *p != '\0')
	    *q++ = *p++;

	*q = '\0';

	if (strlen(tokens[tcnt]) > 0)
	    ++tcnt;
    }

    if (tcnt == 9 || tcnt == 10 || tcnt == 12) {
	n = 0;

	n += sscanf (tokens[0], "%f", &ax);
	n += sscanf (tokens[1], "%f", &ay);
	n += sscanf (tokens[2], "%f", &az);
	n += sscanf (tokens[3], "%f", &bx);
	n += sscanf (tokens[4], "%f", &by);
	n += sscanf (tokens[5], "%f", &bz);
	n += sscanf (tokens[6], "%f", &cx);
	n += sscanf (tokens[7], "%f", &cy);
	n += sscanf (tokens[8], "%f", &cz);

	if (swap_yz) {
	    fswap (&ay, &az);
	    fswap (&by, &bz);
	    fswap (&cy, &cz);
	}

	if (n != 9)
	    return -1;
    }

    if (tcnt == 9)
	return 1;
    else if (tcnt == 12) {
	n = 0;

	n += sscanf (tokens[9], "%f", &red);
	n += sscanf (tokens[10], "%f", &green);
	n += sscanf (tokens[11], "%f", &blue);

	if (n != 3)
	    return -1;

	return 2;
    }
    else if (tcnt == 10) {
	if (!isdigit(tokens[9][0]))
            strcpy (new_texture, tokens[9]);
        else {
            new_texture[0] = '_';
            strcpy (&new_texture[1], tokens[9]);
	}

	/* Replace all illegal charaters in name with underscores */
	for (i = 0; new_texture[i] != '\0'; i++) {
	    if (!isalnum(new_texture[i]))
		new_texture[i] = '_';
        }

	return 3;
    }
    else if (tcnt == 1) {
	if (!isdigit(tokens[0][0]))
	    strcpy (new_obj_name, tokens[0]);
	else {
	    new_obj_name[0] = '_';
	    strcpy (&new_obj_name[1], tokens[0]);
	}

	/* Replace all illegal characters in name with underscores */
	for (i = 0; new_obj_name[i] != '\0'; i++) {
	    if (!isalnum(new_obj_name[i]))
		new_obj_name[i] = '_';
	}

	return 4;
    }
    else
	return -1;
}


void update_txtlist (char *new_texture)
{
    int i;

    for (i = 0; i < txtcnt; i++) {
	if (strcmp (new_texture, txtlist[i]) == 0)
	    break;
    }

    if (i < txtcnt)
	return;

    if (i == MAX_TEXTURE)
	abortmsg ("Too many textures", 1);

    txtlist[i] = malloc (strlen (new_texture) + 1);
    strcpy (txtlist[i], new_texture);

    ++txtcnt;
}


void fswap (float *a, float *b)
{
    float temp;

    temp = *a;
    *a = *b;
    *b = temp;
}
